/*
 * mod_comb - implementation of non-standard COMB command for ProFTPD
 * Copyright (c) 2008 Lev Serebryakov, <lev@serebryakkov.spb.ru>
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*
 * mod_comb module for ProFTPD
 * $Id: mod_comb.c 1965 2008-02-16 19:41:28Z lev $
 */

#include "conf.h"

#define MOD_COMB_VERSION		"mod_comb/0.1"
#define C_COMB					"COMB"

#define COMB_BUF_MINSIZE		65536			/* 64K */
#define COMB_BUF_MAXSIZE		1048576			/*  1M */

#define COMB_NAMECHECK			1

/* Command handlers
 */

/* This command handler is for a non-standard FTP command, "COMB".
 */
MODRET comb_cmd(cmd_rec *cmd) {
  char **paths;
  size_t *sizes;
  config_rec *c = NULL;
  long scount = 0;
  size_t tsize = 0, ssize = 0;
  size_t rsize = 0;
  struct stat st;
  pr_fh_t *fh_in, *fh_out;
  int i;
  char *buf;
  size_t bufsize = 0;
  size_t rd, wr, off;
  int baseLen = 0;
  char ncheck = COMB_NAMECHECK;

  if(cmd->argc < 3) {
    pr_response_add_err(R_501, "COMB command needs at least two arguments");
    return PR_ERROR(cmd);
  }

  /* Check limits */
  c = find_config(main_server->conf, CONF_PARAM, "CombMaxSegmentCount", FALSE);
  if(c) {
    scount = *((long *)c->argv[0]);
  }
  c = find_config(main_server->conf, CONF_PARAM, "CombMaxSegmentSize", FALSE);
  if(c) {
    ssize = *((size_t *)c->argv[0]);
  }
  c = find_config(main_server->conf, CONF_PARAM, "CombMaxTotalSize", FALSE);
  if(c) {
    tsize = *((size_t *)c->argv[0]);
  }
  c = find_config(main_server->conf, CONF_PARAM, "CombCheckSegmentNames", FALSE);
  if(c) {
    ncheck = *((char *)c->argv[0]);
  }
  c = find_config(main_server->conf, CONF_PARAM, "CombBufferSize", FALSE);
  if(c) {
    bufsize = *((size_t *)c->argv[0]);
  }
  if(bufsize < COMB_BUF_MINSIZE) {
    bufsize = COMB_BUF_MINSIZE;
  } else if(bufsize > COMB_BUF_MAXSIZE) {
    bufsize = COMB_BUF_MAXSIZE;
  }
  
  /* Check number of files */
  if(scount && cmd->argc - 1 > scount) {
    pr_response_add_err(R_550, "COMB: too many segments");
    return PR_ERROR(cmd);
  }
  
  /* prepare absolute name and check them */
  paths = pcalloc(cmd->tmp_pool, sizeof(char *) * (cmd->argc - 1));
  sizes = pcalloc(cmd->tmp_pool, sizeof(size_t) * (cmd->argc - 1));

  for(i = 1; i < cmd->argc; ++i) {
    paths[i-1] = dir_best_path(cmd->tmp_pool, pr_fs_decode_path(cmd->tmp_pool, cmd->argv[i]));
    if(!paths[i-1]) {
      pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(errno));
      return PR_ERROR(cmd);
    }
    /* Check TWO commands: this one and DELE for all but first segments */
    if(!dir_check(cmd->tmp_pool, cmd->argv[0], cmd->group, paths[i-1], NULL)) {
      pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(errno));
      return PR_ERROR(cmd);
    }
    /* Additional checks for all but first segments */
    if(i > 1) {
      if(!dir_check(cmd->tmp_pool, C_DELE, cmd->group, paths[i-1], NULL)) {
        pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(errno));
        return PR_ERROR(cmd);
      }
      if(ncheck && 
          (strlen(paths[i-1]) <= baseLen ||
           memcmp(paths[0], paths[i-1], baseLen))) {
        pr_response_add_err(R_550, "%s: Is not valid segment name", cmd->argv[i]);
        return PR_ERROR(cmd);
      }
    } else {
      baseLen = strlen(paths[i-1]);
    }
    
    /* Check existense & get size */
    if(-1 == pr_fsio_stat(paths[i-1], &st)) {
      pr_response_add_err(R_500, "COMB: %s not found", cmd->argv[i]);
      return PR_ERROR(cmd);
    }
    if(ssize && st.st_size > ssize) {
      pr_response_add_err(R_452, "COMB: %s is too large", cmd->argv[i]);
      return PR_ERROR(cmd);
    }
    sizes[i-1] = st.st_size;
    rsize += st.st_size;
  }
  
  /* check total size */
  if(tsize && rsize > tsize) {
    pr_response_add_err(R_452, "COMB: result will be too large");
    return PR_ERROR(cmd);
  }

  pr_fs_setcwd(pr_fs_getcwd());
  
  /* Open first path for write & seek in it */
  fh_out = pr_fsio_open(paths[0], O_WRONLY);
  if(!fh_out) {
    pr_response_add_err(R_550, "%s: %s", cmd->argv[1], strerror(errno));
    return PR_ERROR(cmd);
  }
  if((off_t)-1 == pr_fsio_lseek(fh_out, 0, SEEK_END)) {
    pr_response_add_err(R_550, "%s: can not append", cmd->argv[1]);
    return PR_ERROR(cmd);
  }
  
  buf = pcalloc(cmd->tmp_pool, bufsize);

  for(i = 2; i < cmd->argc; ++i) {
    fh_in = pr_fsio_open(paths[i-1], O_RDONLY);
    if(!fh_out) {
      pr_response_add_err(R_550, "%s: %s", cmd->argv[i], strerror(errno));
      return PR_ERROR(cmd);
    }
    /* Copy out */
    while((rd = pr_fsio_read(fh_in, buf, bufsize)) > 0 && sizes[i-1]) {
      sizes[i-1] -= rd;
      off = 0;
      while((wr = pr_fsio_write(fh_out, buf + off, rd)) > 0 && rd) {
        off += wr;
        rd  -= wr;
      }
      if(rd) {
        pr_response_add_err(R_550, "%s: %s", cmd->argv[1], strerror(errno));
        pr_fsio_close(fh_in);
        pr_fsio_close(fh_out);
        return PR_ERROR(cmd);
      }
    }
    if(sizes[i-1]) {
      pr_response_add_err(R_550, "%s: %s", cmd->argv[i], strerror(errno));
      pr_fsio_close(fh_in);
      pr_fsio_close(fh_out);
      return PR_ERROR(cmd);
      }
    pr_fsio_close(fh_in);
    pr_fsio_unlink(paths[i-1]);
  }
  pr_fsio_close(fh_out);

  pr_response_add(R_250, _("COMB command successful"));
  return PR_HANDLED(cmd);
}

/* Configuration handlers 
 */

MODRET set_max_segment_count(cmd_rec *cmd) {
  long count = 0;
  char *endp;
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_ANON|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR|CONF_DYNDIR);

  count = strtol(cmd->argv[1], &endp, 10);
  if (endp && *endp)
    CONF_ERROR(cmd, "not a valid number");
    
  if(count == 1 || count < 0)
    CONF_ERROR(cmd, "not a valid number of segments");

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(count));
  *((long *) c->argv[0]) = count;
  c->flags |= CF_MERGEDOWN;

  return PR_HANDLED(cmd);
}

MODRET set_max_xxx_size(cmd_rec *cmd) {
  size_t size  = 0;
  unsigned long scale = 1;
  char scale_str[16];
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_ANON|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR|CONF_DYNDIR);
  
  if(2 != sscanf(cmd->argv[1],"%lu%2s", &size, (char*)&scale_str))
    CONF_ERROR(cmd, "not a valid file size");
  
  if       (!scale_str[0] || !strcasecmp(scale_str,"B")) {
    /* No scale: bytes */
  } else if(!strcasecmp(scale_str,"G") || !strcasecmp(scale_str,"Gb")) {
    scale = 1024 * 1024 * 1024;
  } else if(!strcasecmp(scale_str,"M") || !strcasecmp(scale_str,"Mb")) {
    scale = 1024 * 1024;
  } else if(!strcasecmp(scale_str,"K") || !strcasecmp(scale_str,"Kb")) {
    scale = 1024;
  } else {
    CONF_ERROR(cmd, "not a valid file size");
  }
  /* check max size */
  if (size > (SIZE_MAX / scale)) {
    scale = 1;
    size = SIZE_MAX;
  } else {
    size *= scale;
  }

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(size));
  *((size_t *) c->argv[0]) = size;
  c->flags |= CF_MERGEDOWN;

  return PR_HANDLED(cmd);
}

MODRET set_check_name(cmd_rec *cmd) {
  char bool = 0;
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_ANON|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR|CONF_DYNDIR);

  bool = get_boolean(cmd, 1);
  if (bool == -1)
    CONF_ERROR(cmd, "requires a Boolean parameter");
    
  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(bool));
  *((unsigned char *) c->argv[0]) = bool;
  c->flags |= CF_MERGEDOWN;

  return PR_HANDLED(cmd);
}


/* Initialization routines
 */
 
static int comb_init(void) {
  pr_help_add(C_COMB, "<sp> pathname <sp> pathname [<sp> pathname ...]", TRUE);
  return 0;
}

static int comb_init_sess(void) {
  /* We provide COMB feature */
  pr_feat_add("COMB");
  return 0;
}

static conftable comb_conftab[] = {
  { "CombMaxSegmentCount",		set_max_segment_count,	NULL },
  { "CombMaxSegmentSize",		set_max_xxx_size,		NULL },
  { "CombMaxTotalSize",			set_max_xxx_size,		NULL },
  { "CombBufferSize",			set_max_xxx_size,		NULL },
  { "CombCheckSegmentNames",	set_check_name,			NULL },
  { NULL }
};

static cmdtable comb_cmdtab[] = {
  { CMD,	C_COMB,	G_WRITE, comb_cmd, TRUE, FALSE, CL_WRITE|CL_MISC },
  { 0,		NULL }
};

module comb_module = {

  /* Always NULL */
  NULL, NULL,

  /* Module API version (2.0) */
  0x20,

  /* Module name */
  "comb",

  /* Module configuration directive handlers */
  comb_conftab,

  /* Module command handlers */
  comb_cmdtab,

  /* Module authentication handlers (none in this case) */
  NULL,

  /* Module initialization */
  comb_init,

  /* Session initialization */
  comb_init_sess,

  /* Module version */
  MOD_COMB_VERSION
};
